from __future__ import annotations

from dataclasses import dataclass, field
from typing import Self

from ..cell.slice import Slice
from ..cell.builder import Builder

from .hashmap import HashmapE

from .tlb import TlbConstructor

@dataclass(frozen=True, slots=True)
class ExtraCurrencyCollection(TlbConstructor):
    """
    extra_currencies$_ dict:(HashmapE 32 (VarUInteger 32)) = ExtraCurrencyCollection;
    """
    dict: HashmapE = None

    @classmethod
    def tag(cls):
        return None

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_hashmap_e(self.dict)

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        dict = s.load_hashmap_e(32)
        return cls(dict)


@dataclass(frozen=True, slots=True)
class CurrencyCollection(TlbConstructor):
    """
    currencies$_ grams:Grams other:ExtraCurrencyCollection = CurrencyCollection;
    """
    grams: int
    other: ExtraCurrencyCollection

    @classmethod
    def tag(cls):
        return None

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_coins(self.grams)
        b.store_tlb(self.other)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        grams = s.load_coins()
        other = s.load_tlb(ExtraCurrencyCollection)
        return cls(grams, other)



from typing import Self
from dataclasses import dataclass

from ..cell.slice import Slice
from ..cell.cell import Cell
from ..cell.builder import Builder

from .tlb import TlbConstructor

@dataclass(frozen=True, slots=True)
class TickTock(TlbConstructor):
    '''tick_tock$_ tick:Bool tock:Bool = TickTock;'''
    tick: bool
    tock: bool

    @classmethod
    def tag(cls):
        return None

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_bool(self.tick)
        b.store_bool(self.tock)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        tick = s.load_bool()
        tock = s.load_bool()
        return cls(tick, tock)


@dataclass(frozen=True, slots=True)
class StateInit(TlbConstructor):
    '''
    _ fixed_prefix_length:(Maybe (## 5)) special:(Maybe TickTock)
      code:(Maybe ^Cell) data:(Maybe ^Cell)
      library:(Maybe ^Cell) = StateInit;
    '''
    fixed_prefix_length: int | None = None
    special: TickTock | None = None
    code: Cell | None = None
    data: Cell | None = None
    library: Cell | None = None

    @classmethod
    def tag(cls):
        return None

    def serialize_fields(self, b: Builder | None = None) -> Builder:
        if b is None:
            b = Builder()
        if self.fixed_prefix_length is not None:
            b.store_bit(1)
            b.store_uint(self.fixed_prefix_length, 5)
        else:
            b.store_bit(0)
        b.store_maybe_tlb(self.special)
        b.store_maybe_ref(self.code)
        b.store_maybe_ref(self.data)
        b.store_maybe_ref(self.library)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        fixed_prefix_length = s.load_uint(5) if s.load_bit() else None
        special = s.load_tlb(TickTock)
        code = s.load_maybe_ref()
        data = s.load_maybe_ref()
        library = s.load_maybe_ref()
        return cls(fixed_prefix_length, special, code, data, library)



from __future__ import annotations

from typing import Self, Iterable

from ..cell.cell import Cell

from ..provider import Provider
from ..types import MsgAddressInt, Address, StateInit, MessageRelaxed, Message, CurrencyCollection
from ..types import MsgAddressExt, AddrNone
from ..types.tvm_value import TvmValue

class Contract:
    def __init__(self, provider: Provider, address: MsgAddressInt) -> None:
        self.provider = provider
        self.address = address
    
    def run_get_method(self,
                       method_id: int | str,
                       stack: Iterable[TvmValue] | TvmValue | None = None) -> tuple[TvmValue, ...] | TvmValue:
        return self.provider.run_get_method(self.address, method_id, stack)
    
    def create_internal_message(self, *,
                                value: int | CurrencyCollection = 0,
                                body: Cell = Cell.empty(),
                                bounce: bool = True,
                                init: StateInit | None = None) -> MessageRelaxed:
        return MessageRelaxed.internal(
            dest=self.address,
            value=value,
            body=body,
            bounce=bounce,
            init=init
        )
    
    def create_external_message(self,
                                body: Cell = Cell.empty(),
                                src: MsgAddressExt = AddrNone(),
                                init: StateInit | None = None) -> Message:
        return Message.external_in(
            src=src,
            dest=self.address,
            body=body,
            init=init
        )
    
    @staticmethod
    def state_init_to_address(state_init: StateInit, wc: int = 0) -> Address:
        return Address(wc, state_init.to_cell().hash())
    
    @staticmethod
    def code_and_data_to_address(code: Cell, data: Cell, wc: int = 0) -> Address:
        return Contract.state_init_to_address(StateInit(code=code, data=data), wc)



from __future__ import annotations

import json
from collections.abc import Mapping
from typing import Any, Literal, overload, Self

import requests
import requests.auth
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


JSON = dict[str, Any] | list[Any] | str | int | float | bool | None

class ApiError(requests.HTTPError):
    """Raised for non-2xx responses with handy context attached."""

    def __init__(self, message: str, response: requests.Response):
        try:
            payload = response.json()
        except Exception:
            payload = response.text
        self.status_code: int = response.status_code
        self.response: requests.Response = response
        self.payload: Any = payload
        super().__init__(message, response=response)

    def __str__(self) -> str:  # pragma: no cover
        base = f"{self.status_code} {self.response.reason}"
        try:
            detail = json.dumps(self.payload, ensure_ascii=False)[:500]
        except Exception:
            detail = str(self.payload)[:500]
        return f"{base} - {detail}"


class BaseApiClient:
    """
    Minimal base for HTTP API clients using `requests`.
    Subclass and add typed methods that call self.get/post/... with paths like "/v1/things".
    """

    DEFAULT_TIMEOUT: float = 15.0  # seconds

    def __init__(
        self,
        base_url: str,
        *,
        headers: Mapping[str, str] | None = None,
        bearer_token: str | None = None,
        api_key: tuple[str, str] | None = None,  # (header_name, value)
        auth: requests.auth.AuthBase | None = None,
        timeout: float | None = None,
        retries: int | None = 3,
        backoff_factor: float = 0.4,
        status_forcelist: tuple[int, ...] = (429, 500, 502, 503, 504),
        allowed_methods: tuple[str, ...] = ("HEAD", "GET", "PUT", "POST", "DELETE", "OPTIONS", "PATCH"),
        proxies: Mapping[str, str] | None = None,
        verify: bool | str = True,  # path to CA bundle or bool
        user_agent: str = "BaseApiClient/2.0 (+https://example.com)",
    ) -> None:
        if not base_url.endswith("/"):
            base_url += "/"
        self.base_url = base_url
        self.verify = verify
        self.proxies = dict(proxies) if proxies else None
        self.timeout = float(timeout) if timeout is not None else self.DEFAULT_TIMEOUT

        sess = requests.Session()

        # Retries with exponential backoff, including on 429 rate limits.
        if retries and retries > 0:
            retry = Retry(
                total=retries,
                connect=retries,
                read=retries,
                redirect=retries,
                status=retries,
                backoff_factor=backoff_factor,
                status_forcelist=frozenset(status_forcelist),
                allowed_methods=frozenset(m.upper() for m in allowed_methods),
                raise_on_status=False,
            )
            adapter = HTTPAdapter(max_retries=retry, pool_connections=100, pool_maxsize=100)
            sess.mount("https://", adapter)
            sess.mount("http://", adapter)

        # Default headers
        self._default_headers: dict[str, str] = {
            "User-Agent": user_agent,
            "Accept": "application/json, */*;q=0.1",
        }
        if headers:
            self._default_headers.update(headers)

        # Auth
        self._auth = auth
        if bearer_token:
            self._default_headers["Authorization"] = f"Bearer {bearer_token}"
        if api_key:
            key_header, key_value = api_key
            self._default_headers[key_header] = key_value

        self.session = sess

    # ---- lifecycle --------------------------------------------------------

    def close(self) -> None:
        self.session.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    # ---- verb helpers -----------------------------------------------------

    def get(self, path: str, **kwargs) -> Any:
        return self.request("GET", path, **kwargs)

    def post(self, path: str, **kwargs) -> Any:
        return self.request("POST", path, **kwargs)

    def put(self, path: str, **kwargs) -> Any:
        return self.request("PUT", path, **kwargs)

    def patch(self, path: str, **kwargs) -> Any:
        return self.request("PATCH", path, **kwargs)

    def delete(self, path: str, **kwargs) -> Any:
        return self.request("DELETE", path, **kwargs)

    # ---- core request -----------------------------------------------------

    @overload
    def request(
        self,
        method: str,
        path: str,
        *,
        params: Mapping[str, Any] | None = ...,
        headers: Mapping[str, str] | None = ...,
        json: JSON = ...,
        data: Mapping[str, Any] | bytes | None = ...,
        files: Mapping[str, Any] | None = ...,
        timeout: float | None = ...,
        auth: requests.auth.AuthBase | None = ...,
        stream: Literal[True] = ...,
        allow_redirects: bool = ...,
        absolute_url: str | None = ...,
    ) -> requests.Response: ...
    @overload
    def request(
        self,
        method: str,
        path: str,
        *,
        params: Mapping[str, Any] | None = ...,
        headers: Mapping[str, str] | None = ...,
        json: JSON = ...,
        data: Mapping[str, Any] | bytes | None = ...,
        files: Mapping[str, Any] | None = ...,
        timeout: float | None = ...,
        auth: requests.auth.AuthBase | None = ...,
        stream: Literal[False] = ...,
        allow_redirects: bool = ...,
        absolute_url: str | None = ...,
    ) -> Any: ...
    def request(
        self,
        method: str,
        path: str,
        *,
        params: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
        json: JSON = None,
        data: Mapping[str, Any] | bytes | None = None,
        files: Mapping[str, Any] | None = None,
        timeout: float | None = None,
        auth: requests.auth.AuthBase | None = None,
        stream: bool = False,
        allow_redirects: bool = True,
        absolute_url: str | None = None,
    ) -> Any:
        """
        Core request method. Usually pass a relative `path` like "/v1/items".
        If you already have a full URL, pass it via `absolute_url` (path is ignored).
        """
        url = self._build_url(path) if not absolute_url else absolute_url

        req_headers = self._merge_headers(headers)

        resp = self.session.request(
            method=method.upper(),
            url=url,
            params=params,
            headers=req_headers,
            json=json,
            data=data,
            files=files,
            timeout=self._coerce_timeout(timeout),
            auth=auth if auth is not None else self._auth,
            stream=stream,
            allow_redirects=allow_redirects,
            verify=self.verify,
            proxies=self.proxies,
        )

        if stream:  # caller handles streaming/closing
            self._raise_for_status(resp)
            return resp

        return self._handle_response(resp)

    # ---- internals --------------------------------------------------------

    def _build_url(self, path: str) -> str:
        if path.startswith(("http://", "https://")):
            return path
        # allow paths with/without leading slash
        return f"{self.base_url}{path.lstrip('/')}"

    def _merge_headers(self, headers: Mapping[str, str] | None) -> dict[str, str]:
        merged = dict(self._default_headers)
        if headers:
            merged.update(headers)
        return merged

    def _coerce_timeout(self, timeout: float | None) -> float:
        return float(timeout) if timeout is not None else self.timeout

    def _handle_response(self, resp: requests.Response) -> Any:
        self._raise_for_status(resp)

        ctype = resp.headers.get("Content-Type", "")
        if "json" in ctype:
            try:
                return resp.json()
            except json.JSONDecodeError:
                pass  # fall through to text

        if "text/" in ctype or not ctype:
            return resp.text
        return resp.content

    def _raise_for_status(self, resp: requests.Response) -> None:
        if 200 <= resp.status_code < 300:
            return
        method = resp.request.method if resp.request else "REQUEST"
        url = resp.url
        msg = f"{method} {url} -> {resp.status_code}"
        raise ApiError(msg, resp)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Self, Iterable, TYPE_CHECKING
from bitarray import bitarray, frozenbitarray
from bitarray.util import ba2int
from hashlib import sha256

from agton.ton.common.bitstring import BitString

from ..crypto.signing import sign, verify

if TYPE_CHECKING:
    from .slice import Slice
    from .builder import Builder

MAX_BITS: int = 1023
MAX_REFS: int = 4

class CellType(Enum):
    ordinary = -1
    pruned_branch = 1
    library_ref = 2
    merkle_proof = 3
    merkle_update = 4


class LevelMask(int):
    def __new__(cls, value: int):
        return super().__new__(cls, value)
    
    def level(self) -> int:
        return self.bit_length()
    
    def hash_index(self) -> int:
        return self.bit_count()
    
    def hashes_count(self) -> int:
        return self.hash_index() + 1
    
    def apply(self, level: int) -> LevelMask:
        return LevelMask(self & ((1 << level) - 1))
    
    def is_significant(self, level: int) -> bool:
        if level == 0:
            return True
        return bool(self >> (level - 1) & 1)


class Cell:
    def __init__(self,
                 data: BitString,
                 refs: Iterable[Self] = (),
                 special: bool = False) -> None:
        self.data = data
        self.refs = tuple(refs)
        self.special = special

        self.type = self._get_type_and_check()
        self.level_mask = self._get_mask()

        self._depthes: list[int] = []
        self._hashes: list[bytes] = []
        self._calculate_hashes_and_depthes()

    def _get_type_and_check(self) -> CellType:
        if not self.special:
            if len(self.refs) > MAX_REFS:
                raise ValueError(f"Ordinary cell can have at most {MAX_REFS} references, got {len(self.refs)}")
            if len(self.data) > MAX_BITS:
                raise ValueError(f"Ordinary cell can have at most {MAX_BITS} bits in data, got {len(self.data)}")
            return CellType.ordinary
        else:
            if len(self.data) < 8:
                raise ValueError(f"Special cell must have at least 8 bits in data")
            t = ba2int(self.data[:8])
            if t == CellType.pruned_branch.value:
                m = ba2int(self.data[8:16])
                h = m.bit_count()
                if not (1 <= m <= 7 and 
                        len(self.refs) == 0 and 
                        len(self.data) == 16 + 8 * (h * 32 + h * 2)):
                    raise ValueError(f"Pruned branch is invalid")
                return CellType.pruned_branch
            elif t == CellType.library_ref.value:
                if not (len(self.refs) == 0 and len(self.data) == 8 + 256):
                    raise ValueError(f"Library cell is invalid")
                return CellType.library_ref
            elif t == CellType.merkle_proof.value:
                if not (len(self.refs) == 1 and len(self.data) == 280):
                    raise ValueError(f"Merkle proof cell is invalid")
                return CellType.merkle_proof
            elif t == CellType.merkle_update.value:
                if not (len(self.refs) == 2 and len(self.data) == 552):
                    raise ValueError(f"Merkle update cell is invalid")
                return CellType.merkle_update
            else:
                raise ValueError(f"Unknown cell type: {t}")
    
    def _get_mask(self) -> LevelMask:
        if self.type == CellType.ordinary:
            m = 0
            for c in self.refs:
                m |= c.level_mask
            return LevelMask(m)
        elif self.type == CellType.pruned_branch:
            return LevelMask(ba2int(self.data[8:16]))
        elif self.type == CellType.merkle_proof:
            return LevelMask(self.refs[0].level_mask >> 1)
        elif self.type == CellType.merkle_update:
            return LevelMask((self.refs[0].level_mask | self.refs[1].level_mask) >> 1)
        elif self.type == CellType.library_ref:
            return LevelMask(0)
        else:
            assert False
    
    def _get_refs_descriptor(self, level_mask: LevelMask) -> bytes:
        d = len(self.refs) + 8 * self.special + 32 * level_mask.level()
        return d.to_bytes(1, 'big')

    def _get_bits_descriptor(self) -> bytes:
        b = len(self.data)
        d = b // 8 + (b + 7) // 8
        return d.to_bytes(1, 'big')

    def _get_descriptors(self, level_mask: LevelMask = LevelMask(0)) -> bytes:
        return self._get_refs_descriptor(level_mask) + self._get_bits_descriptor()

    def depth(self, level: int | None = None) -> int:
        if level is None:
            level = self.level_mask.level()
        hash_index = self.level_mask.apply(level).hash_index()
        if self.type == CellType.pruned_branch:
            pruned_hash_index = self.level_mask.hash_index()
            if hash_index != pruned_hash_index:
                off = 2 + 32 * pruned_hash_index + hash_index * 2
                return int.from_bytes(self._get_data_bytes()[off: off + 2], 'big')
            hash_index = 0
        return self._depthes[hash_index]

    def hash(self, level: int | None = None) -> bytes:
        if level is None:
            level = self.level_mask.level()
        hash_index = self.level_mask.apply(level).hash_index()
        if self.type == CellType.pruned_branch:
            pruned_hash_index = self.level_mask.hash_index()
            if hash_index != pruned_hash_index:
                return self._get_data_bytes()[2 + (hash_index * 32): 2 + ((hash_index + 1) * 32)]
            hash_index = 0
        return self._hashes[hash_index]
    
    def _calculate_hashes_and_depthes(self) -> None:
        total_hash_count = self.level_mask.hashes_count()
        hash_count = total_hash_count
        if self.type == CellType.pruned_branch:
            hash_count = 1
        hash_index_offset = total_hash_count - hash_count
        hash_index = 0
        level = self.level_mask.level()
        for li in range(0, level + 1):
            if not self.level_mask.is_significant(li):
                continue
            if li < hash_index_offset:
                hash_index += 1
                continue
            dsc = self._get_descriptors(self.level_mask.apply(li))
            hasher = sha256(dsc)
            if hash_index == hash_index_offset:
                if li != 0 and self.type != CellType.pruned_branch:
                    raise ValueError('not pruned or 0')
                data = self._get_data_bytes()
                hasher.update(data)
            else:
                if li == 0 or self.type == CellType.pruned_branch:
                    raise ValueError('not pruned or 0')
                off = hash_index - hash_index_offset - 1
                hasher.update(self._hashes[off])
            depth = 0
            for r in self.refs:
                if self.type in (CellType.merkle_proof, CellType.merkle_update):
                    ref_depth = r.depth(li + 1)
                else:
                    ref_depth = r.depth(li)
                depth_bytes = ref_depth.to_bytes(2, 'big')
                hasher.update(depth_bytes)
                depth = max(depth, ref_depth + 1)

            if depth >= 1024:
                raise ValueError('depth is more than max depth')
            for r in self.refs:
                if self.type in (CellType.merkle_proof, CellType.merkle_update):
                    hasher.update(r.hash(li + 1))
                else:
                    hasher.update(r.hash(li))
            off = hash_index - hash_index_offset
            self._depthes.append(depth)
            self._hashes.append(hasher.digest())
            hash_index += 1

    def _get_data_bytes(self) -> bytes:
        result = bitarray(self.data)
        if len(result) % 8:
            result.append(1)
            result.fill()
        return result.tobytes()

    @classmethod
    def empty(cls) -> Cell:
        return cls(data=frozenbitarray(), refs=(), special=False)

    def to_boc(self, 
               with_crc=False,
               with_index=False,
               with_cache=False,
               with_top_hash=False,
               with_int_hashes=False) -> bytes:
        from .boc import encode
        return encode([self], with_crc, with_index, with_cache, with_top_hash, with_int_hashes)
    
    @classmethod
    def from_boc(cls, b: bytes | str) -> Cell:
        from .boc import decode
        if isinstance(b, str):
            b = bytes.fromhex(b)
        roots = decode(b)
        if len(roots) != 1:
            raise ValueError(f'Expected exactly one root in BoC, but {len(roots)} found')
        return roots[0]
    
    def begin_parse(self, allow_exotic: bool = False) -> Slice:
        from .slice import Slice
        if not allow_exotic and self.special:
            raise ValueError('Cannot parse exotic cell, use allow_exotic=True if you want to parse exotic internals')
        return Slice(self, 0, len(self.data), 0, len(self.refs))
    
    def to_slice(self, allow_exotic: bool = False):
        return self.begin_parse(allow_exotic)
    
    def to_builder(self) -> Builder:
        from .builder import Builder
        if self.special:
            raise ValueError("Can't convert special cell to builder")
        b = Builder()
        b.data = bitarray(self.data)
        b.refs = list(self.refs)
        return b
    
    def sign(self, private_key: bytes) -> bytes:
        return sign(self.hash(), private_key)
    
    def verify(self, signature: bytes, public_key: bytes) -> bool:
        return verify(self.hash(), signature, public_key)

    def _type_description(self):
        desc = ''
        if self.type == CellType.pruned_branch:
            desc = '* Prunned Branch '
        elif self.type == CellType.merkle_proof:
            desc = '* Merkle Proof '
        elif self.type == CellType.merkle_update:
            desc = '* Merkle Update '
        elif self.type == CellType.library_ref:
            desc = '* Lirary Ref '
        return desc

    def dump(self, d: int = 0, comma: bool = False) -> str:
        desc = self._type_description()
        data = f'{len(self.data)}[{self.data.tobytes().hex().upper()}]'
        refs = ''
        if self.refs:
            refs = ' -> {\n'
            for c in self.refs[:-1]:
                refs += '\t' * d + '\t' + c.dump(d + 1, comma=True) + '\n'
            refs += '\t' * d + '\t' + self.refs[-1].dump(d + 1) + '\n'
            refs += '\t' * d + '}'
        return desc + data + refs + (',' if comma else '')

    def __hash__(self) -> int:
        return hash(self.hash())
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Cell):
            return False
        return self.hash() == other.hash()

    def __repr__(self) -> str:
        desc = self._type_description()
        return f"Cell({desc}{len(self.data)}[{self.data.tobytes().hex().upper()}] -> {len(self.refs)} refs)"

    def __str__(self) -> str:
        return self.dump()



from __future__ import annotations

from bitarray import frozenbitarray
from bitarray.util import ba2int
from typing import Self, Literal, Callable, cast, TYPE_CHECKING

from agton.ton.common.bitstring import BitString


from .cell import Cell
from .exceptions import CellOverflow, CellUnderflow

if TYPE_CHECKING:
    from ..types.tlb import TlbConstructor
    from ..types.msg_address import Address, MsgAddress, MsgAddressExt, MsgAddressInt
    from ..types.hashmap import Hashmap, HashmapE


class Slice:
    def __init__(self, c: Cell, ld: int, rd: int, lr: int, rr: int) -> None:
        self.c = c
        self.ld = ld
        self.rd = rd
        self.lr = lr
        self.rr = rr
        if not (ld <= rd <= len(c.data) and lr <= rr <= len(c.refs)):
            raise ValueError('Invalid cursors')
    
    def to_cell(self) -> Cell:
        return Cell(
            data=frozenbitarray(self.c.data[self.ld:self.rd]),
            refs=self.c.refs[self.lr:self.ld],
            special=False
        )
    
    @property
    def remaining_bits(self):
        return self.rd - self.ld
    
    @property
    def remaining_refs(self):
        return self.rr - self.lr

    def _ensure_bits_cap(self, n: int):
        if n > self.remaining_bits:
            raise CellUnderflow(f'Tried to read {n} bits, when {self.remaining_bits} available')
    
    def _ensure_refs_cap(self, n: int):
        if n > self.remaining_refs:
            raise CellUnderflow(f'Tried to read {n} refs, when {self.remaining_refs} available')
    
    def end_parse(self):
        if not (self.lr == self.rr and self.ld == self.rd):
            raise CellOverflow('Expected no more data and refs, after .end_parse()')
    
    def with_skipped_bits(self, n: int) -> Slice:
        self._ensure_bits_cap(n)
        return Slice(self.c, self.ld + n, self.rd, self.lr, self.rr)
    
    def with_skipped_refs(self, n: int) -> Slice:
        self._ensure_refs_cap(n)
        return Slice(self.c, self.ld, self.rd, self.lr + n, self.rr)

    def preload_bits(self, n: int) -> BitString:
        self._ensure_bits_cap(n)
        return BitString(self.c.data[self.ld:self.ld + n])
    
    def load_bits(self, n: int) -> BitString:
        self._ensure_bits_cap(n)
        self.ld += n
        return BitString(self.c.data[self.ld - n:self.ld])
    
    def starts_with(self, b: BitString) -> bool:
        n = len(b)
        return self.preload_bits(n) == b

    def skip_bits(self, n: int) -> Self:
        self._ensure_bits_cap(n)
        self.ld += n
        return self

    def skip_prefix(self, b: BitString):
        n = len(b)
        if not self.starts_with(b):
            raise ValueError('Wrong prefix')
        self.skip_bits(n)

    def preload_ref(self) -> Cell:
        self._ensure_refs_cap(1)
        return self.c.refs[self.lr]
    
    def load_ref(self) -> Cell:
        self._ensure_refs_cap(1)
        self.lr += 1
        return self.c.refs[self.lr - 1]
    
    def preload_bytes(self, n: int) -> bytes:
        return self.preload_bits(n * 8).tobytes()
    
    def load_bytes(self, n: int) -> bytes:
        return self.load_bits(n * 8).tobytes()

    def load_uint(self, n: int) -> int:
        return ba2int(self.load_bits(n), signed=False)
    
    def preload_uint(self, n: int) -> int:
        return ba2int(self.preload_bits(n), signed=False)
    
    def load_int(self, n: int) -> int:
        return ba2int(self.load_bits(n), signed=True)
    
    def preload_int(self, n: int) -> int:
        return ba2int(self.preload_bits(n), signed=True)
    
    def load_bool(self) -> bool:
        return bool(self.load_uint(1))
    
    def preload_bool(self) -> bool:
        return bool(self.preload_uint(1))
    
    def load_bit(self) -> Literal[0, 1]:
        return 1 if self.load_bool() else 0
    
    def preload_bit(self) -> Literal[0, 1]:
        return 1 if self.preload_bool() else 0
    
    def preload_var_uint(self, n: int) -> int:
        bit_length = (n - 1).bit_length()
        length = self.preload_uint(bit_length)
        if not length:
            return 0
        return self.with_skipped_bits(bit_length).load_uint(length * 8)
    
    def load_var_uint(self, n: int) -> int:
        bit_length = (n - 1).bit_length()
        length = self.load_uint(bit_length)
        if not length:
            return 0
        return self.load_uint(length * 8)
    
    def preload_var_int(self, n: int) -> int:
        bit_length = (n - 1).bit_length()
        length = self.preload_uint(bit_length)
        if not length:
            return 0
        return self.with_skipped_bits(bit_length).load_int(length * 8)
    
    def load_var_int(self, n: int) -> int:
        bit_length = (n - 1).bit_length()
        length = self.load_uint(bit_length)
        if not length:
            return 0
        return self.load_int(length * 8)
    
    def load_coins(self) -> int:
        return self.load_var_uint(16)
    
    def preload_coins(self) -> int:
        return self.preload_var_uint(16)

    def load_maybe_ref(self) -> Cell | None:
        if self.load_bool():
            return self.load_ref()
        return None
    
    def load_snake_bytes(self) -> bytes:
        if self.remaining_bits % 8 != 0:
            raise ValueError(f'Invalid string length: {self.remaining_bits}')
        res = self.load_bytes(self.remaining_bits // 8)
        if not self.remaining_refs:
            res += self.load_ref().begin_parse().load_snake_bytes()
        return res

    def load_snake_string(self) -> str:
        return self.load_snake_bytes().decode()
    
    def load_hashmap(self, n: int) -> Hashmap:
        from ..types.hashmap import load_hashmap
        return load_hashmap(self, n)
    
    def load_hashmap_e(self, n: int) -> HashmapE:
        c = self.load_maybe_ref()
        if c is None:
            return None
        return c.begin_parse().load_hashmap(n)
    
    # TlbTypeDeserializer = (
    #     type[TlbConstructor]
    #     | Callable[["Slice"], tuple[TlbConstructor, *tuple[Any, ...]]]
    #     | Callable[["Slice"], TlbConstructor]
    # )

    def load_tlb[T: TlbConstructor](self, deserializer: type[T] | Callable[[Slice], T]) -> T:
        if isinstance(deserializer, type):
            cls = cast(type[T], deserializer)
            return cls.deserialize(self)
        else:
            t = deserializer(self)
            return t
    
    def load_maybe_tlb[T: TlbConstructor](self, deserializer: type[T] | Callable[[Slice], T]) -> T | None:
        if self.load_bool():
            return self.load_tlb(deserializer)
        return None
    
    def load_ref_tlb[T: TlbConstructor](self, deserializer: type[T] | Callable[[Slice], T]) -> T:
        cs = self.load_ref().begin_parse()
        return cs.load_tlb(deserializer)
    
    def load_maybe_ref_tlb[T: TlbConstructor](self, deserializer: type[T] | Callable[[Slice], T]) -> T | None:
        c = self.load_maybe_ref()
        if c is None:
            return None
        cs = c.begin_parse()
        return cs.load_tlb(deserializer)
    
    def load_address(self) -> Address:
        from ..types.msg_address import Address
        return self.load_tlb(Address)
    
    def load_msg_address(self) -> MsgAddress:
        from ..types.msg_address import msg_address
        return self.load_tlb(msg_address)
    
    def load_msg_address_ext(self) -> MsgAddressExt:
        from ..types.msg_address import msg_address_ext
        return self.load_tlb(msg_address_ext)
    
    def load_msg_address_int(self) -> MsgAddressInt:
        from ..types.msg_address import msg_address_int
        return self.load_tlb(msg_address_int)
    
    def load_cell(self) -> Cell:
        '''Use this to load the remaining bits and refs as Cell'''
        c = Cell(frozenbitarray(self.c.data[self.ld:self.rd]), self.c.refs[self.lr:self.rr])
        self.ld = self.rd
        self.lr = self.rr
        return c
    
    def load_slice(self) -> Slice:
        '''Use this to load the remaining bits and refs as Slice'''
        return self.load_cell().begin_parse()
        
    def __repr__(self) -> str:
        data = f'{self.remaining_bits}[{self.c.data[self.ld:self.rd].tobytes().hex().upper()}]'
        return f'Slice({data} -> {self.remaining_refs} refs)'
    
    def __str__(self) -> str:
        c = self.to_cell()
        return str(c)



from __future__ import annotations

from typing import Self, Iterable, overload, TYPE_CHECKING

from bitarray import bitarray, frozenbitarray
from bitarray.util import int2ba

from .cell import Cell
from .exceptions import CellOverflow

if TYPE_CHECKING:
    from ..types.tlb import TlbConstructor
    from ..types.msg_address import Address, MsgAddress, MsgAddressInt, MsgAddressExt
    from ..types.hashmap import HashmapE, Hashmap
    from .slice import Slice


MAX_BITS: int = 1023
MAX_REFS: int = 4

class Builder:
    def __init__(self) -> None:
        self.data = bitarray()
        self.refs = []
    
    @property
    def remaining_bits(self):
        return MAX_BITS - len(self.data)
    
    @property
    def remaining_refs(self):
        return MAX_REFS - len(self.refs)
    
    def end_cell(self) -> Cell:
        return Cell(frozenbitarray(self.data), self.refs)
    
    def to_cell(self) -> Cell:
        return self.end_cell()
    
    def to_slice(self) -> Slice:
        return self.to_cell().to_slice()
    
    def _ensure_bits_cap(self, n: int):
        if n > self.remaining_bits:
            raise CellOverflow(f"Can't store {n} more bits, only {self.remaining_bits} bits available")
    
    def _ensure_refs_cap(self, n: int):
        if n > self.remaining_refs:
            raise CellOverflow(f"Can't store {n} more refs, only {self.remaining_refs} refs available")
    
    def store_ref(self, c: Cell) -> Self:
        self._ensure_refs_cap(1)
        self.refs.append(c)
        return self
    
    def store_bit(self, b: int | bool | str) -> Self:
        if isinstance(b, str):
            if b not in ('0', '1'):
                raise ValueError("Expected only 0 and 1 as str argument")
            b = int(b)
        if isinstance(b, int):
            if b not in (0, 1):
                raise ValueError("Expected only 0 or 1 as int argument")
            b = bool(b)
        return self.store_bool(b)

    def store_bits(self, b: bitarray | str | Iterable[int]) -> Self:
        if not isinstance(b, bitarray):
            b = bitarray(b)
        self._ensure_bits_cap(len(b))
        self.data.extend(b)
        return self

    def store_bytes(self, b: bytes) -> Self:
        n = len(b) * 8
        self._ensure_bits_cap(n)
        self.data.frombytes(b)
        return self
    
    def store_uint(self, v: int, n: int) -> Self:
        b = int2ba(v, n, signed=False)
        self.store_bits(b)
        return self
    
    def store_int(self, v: int, n: int) -> Self:
        b = int2ba(v, n, signed=True)
        self.store_bits(b)
        return self
    
    def store_bool(self, b: bool) -> Self:
        return self.store_uint(b, 1)
    
    def store_var_uint(self, v: int, n: int):
        bit_length = (n - 1).bit_length()
        if v == 0:
            return self.store_uint(0, bit_length)
        byte_length = (v.bit_length() + 7) // 8
        return self.store_uint(byte_length, bit_length).store_uint(v, byte_length * 8)

    def store_var_int(self, v: int, n: int):
        bit_length = (n - 1).bit_length()
        if v == 0:
            return self.store_uint(0, bit_length)
        byte_length = (v.bit_length() + 7) // 8
        return self.store_uint(byte_length, bit_length).store_int(v, byte_length * 8)
    
    def store_coins(self, amount: int):
        return self.store_var_uint(amount, 16)
    
    def store_snake_bytes(self, b: bytes) -> Self:
        n = min(len(b), self.remaining_bits // 8)
        self.store_bytes(b[:n])
        b = b[n:]
        if b:
            self.store_ref(begin_cell().store_snake_bytes(b).end_cell())
        return self
    
    def store_snake_string(self, s: str) -> Self:
        return self.store_snake_bytes(s.encode())
    
    def store_hashmap(self, h: Hashmap) -> Self:
        from ..types.hashmap import store_hashmap
        store_hashmap(self, h)
        return self
    
    def store_hashmap_e(self, h: HashmapE) -> Self:
        if h is None:
            self.store_bit(0)
            return self
        return self.store_hashmap(h)

    def store_builder(self, b: Builder) -> Self:
        self.store_bits(b.data)
        for c in b.refs:
            self.store_ref(c)
        return self
    
    def store_tlb(self, o: TlbConstructor) -> Builder:
        return o.serialize(self)
    
    def store_maybe_tlb(self, o: TlbConstructor | None) -> Builder:
        if o is None:
            return self.store_bit(0)
        self.store_bit(1)
        return o.serialize(self)
    
    def store_ref_tlb(self, o: TlbConstructor) -> Builder:
        return self.store_ref(o.to_cell())
    
    def store_maybe_ref_tlb(self, o: TlbConstructor | None) -> Builder:
        return self.store_maybe_ref(None if o is None else o.to_cell())
    
    def store_address(self, a: Address) -> Builder:
        return self.store_tlb(a)
    
    def store_msg_address(self, a: MsgAddress) -> Builder:
        return self.store_tlb(a)
    
    def store_msg_address_int(self, a: MsgAddressInt) -> Builder:
        return self.store_tlb(a)
    
    def store_msg_address_ext(self, a: MsgAddressExt) -> Builder:
        return self.store_tlb(a)
    
    def store_maybe_ref(self, c: Cell | None) -> Builder:
        if c is None:
            return self.store_bit(0)
        self.store_bit(1)
        return self.store_ref(c)
    
    def store_slice(self, s: Slice) -> Self:
        b = s.to_cell().to_builder()
        return self.store_builder(b)

def begin_cell() -> Builder:
    return Builder()



from agton.ton import Cell, Slice, Builder
from agton.ton.types.continuation import Continuation

TvmValue = (
    None 
    | int 
    | Cell 
    | Slice 
    | Builder
    | Continuation
    | tuple["TvmValue", ...]
)



from typing import Sequence, Iterable

from bitarray import frozenbitarray, bitarray

from agton.ton import Cell
from agton.ton.crypto import crc32c
from agton.ton.common import BytesParser

BOC_MAGIC = b'\xb5\xee\x9c\x72'

def encode(roots: Sequence[Cell],
           with_crc=False,
           with_index=False,
           with_cache=False,
           with_top_hash=False,
           with_int_hashes=False) -> bytes:
    if with_crc or with_index or with_cache or with_top_hash or with_int_hashes:
        raise NotImplementedError

    index: dict[Cell, int] = dict()
    order: list[Cell] = []
    def dfs(c: Cell):
        if c in index:
            return
        index[c] = len(index)
        for r in c.refs:
            dfs(r)
        order.append(c)
    for c in roots:
        dfs(c)

    n = len(index)

    cell_size_bits = n.bit_length()
    cell_size_bytes = (cell_size_bits + 7) // 8

    payload = bytearray()
    for c in reversed(order):
        payload.extend(c._get_descriptors())
        payload.extend(c._get_data_bytes())
        for r in c.refs:
            payload.extend(index[r].to_bytes(cell_size_bytes, 'big'))
    size_bits = len(payload).bit_length()
    size_bytes = (size_bits + 7) // 8
    
    flags = 0
    flags |= cell_size_bytes

    boc = bytearray()
    boc.extend(BOC_MAGIC)
    boc.append(flags)
    boc.append(size_bytes)
    boc.extend(n.to_bytes(cell_size_bytes, 'big'))
    boc.extend(len(roots).to_bytes(cell_size_bytes, 'big'))
    boc.extend((0).to_bytes(cell_size_bytes, 'big'))
    boc.extend(len(payload).to_bytes(size_bytes, 'big'))
    for r in roots:
        boc.extend(index[r].to_bytes(cell_size_bytes, 'big'))
    boc.extend(payload)
    return bytes(boc)

class UnresolvedCell:
    def __init__(self, data: frozenbitarray, refs: Iterable[int], special: bool):
        self.data = data
        self.refs: tuple[int, ...] = tuple(refs)
        self.special = special
    
    def resolve(self, refs: Iterable[Cell]) -> Cell:
        refs = tuple(refs)
        if len(self.refs) != len(refs):
            raise ValueError(f'Invalid resolving, expected {len(self.refs)} refs, but {len(refs)} given')
        return Cell(self.data, refs, self.special)

def decode(boc: bytes) -> list[Cell]:
    parser = BytesParser(boc)
    parser.expect(BOC_MAGIC)
    flags = parser.load_uint(1)
    cell_size_bytes = flags & 7
    has_index = flags >> 7 & 1
    has_crc32c = flags >> 6 & 1
    has_cache_bits = flags >> 5 & 1

    if has_index or has_cache_bits:
        raise NotImplementedError
    
    if has_crc32c:
        crc = crc32c(boc[:-4])
        expected = boc[-4:]
        if crc != expected:
            raise ValueError('Checksum does not match')
    size_bytes = parser.load_uint(1)
    cells_len = parser.load_uint(cell_size_bytes)
    roots_len = parser.load_uint(cell_size_bytes)
    absent = parser.load_uint(cell_size_bytes)
    if absent:
        raise NotImplementedError
    payload_size = parser.load_uint(size_bytes)
    roots_indexes: list[int] = []
    for i in range(roots_len):
        root_index = parser.load_uint(cell_size_bytes)
        roots_indexes.append(root_index)
    payload = parser.load_bytes(payload_size)
    if has_crc32c:
        parser.skip_bytes(4)
    parser.end_parse()

    unresolved_cells: list[UnresolvedCell] = []
    parser = BytesParser(payload)
    for i in range(cells_len):
        d1 = parser.load_uint(1)
        d2 = parser.load_uint(1)
        refs_len = d1 & 7
        special = d1 >> 3 & 1
        has_hashes = d1 >> 4 & 1
        level = d1 >> 5
        if refs_len > 4:
            raise ValueError(f'Too many refs in cell: {refs_len}')
        hashes_size = (level + 1) * 32 if has_hashes else 0
        depth_size = (level + 1) * 2 if has_hashes else 0
        if has_hashes:
            parser.skip_bytes(hashes_size + depth_size)

        is_padded = d2 & 1
        data_size = (d2 >> 1) + is_padded

        data_serialized = parser.load_bytes(data_size)
        data = bitarray()
        data.frombytes(data_serialized)

        if is_padded:
            i = len(data) - 1
            while i >= 0 and not data[i]:
                i -= 1
            data = data[:i]
        
        refs_indexes: list[int] = []
        for i in range(refs_len):
            refs_indexes.append(parser.load_uint(cell_size_bytes))
        unresolved_cells.append(UnresolvedCell(frozenbitarray(data), refs_indexes, bool(special)))
    parser.end_parse()

    cells: list[Cell | None] = [None] * cells_len
    for i, uc in reversed(list(enumerate(unresolved_cells))):
        refs: list[Cell] = []
        for r in uc.refs:
            t = cells[r]
            if t is None:
                raise ValueError('Topological order is broken')
            refs.append(t)
        cells[i] = uc.resolve(refs)
    
    roots: list[Cell] = []
    for i in roots_indexes:
        c = cells[i]
        if c is None:
            raise ValueError(f'Root with index {i} not found in BoC')
        roots.append(c)
    return roots



from __future__ import annotations

from dataclasses import dataclass
from typing import Self, Type

from ..cell.slice import Slice
from ..cell.cell import Cell
from ..cell.builder import Builder, begin_cell

from .tlb import TlbConstructor

from .msg_address import MsgAddressInt, MsgAddressExt, AddrNone
from .state_init import StateInit
from .common_msg_info import CommonMsgInfo, IntMsgInfo, ExtInInfo, ExtOutInfo
from .common_msg_info import common_msg_info
from .currency_collection import CurrencyCollection, ExtraCurrencyCollection

@dataclass(frozen=True, slots=True)
class Message(TlbConstructor):
    '''
    message$_ {X:Type} info:CommonMsgInfo
        init:(Maybe (Either StateInit ^StateInit))
        body:(Either X ^X) = Message X;
    '''
    info: CommonMsgInfo
    init: StateInit | None
    body: Cell

    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Message:
        info = s.load_tlb(common_msg_info)
        init: StateInit | None
        has_init = s.load_bit()
        if has_init:
            init_in_ref = s.load_bit()
            if init_in_ref:
                init = s.load_tlb(StateInit)
            else:
                init = StateInit.from_cell(s.load_ref())
        else:
            init = None
        body: Cell = s.to_cell()
        body_in_ref = s.load_bit()
        if body_in_ref:
            body = s.load_ref()
        return cls(info, init, body)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_tlb(self.info)
        if self.init is None:
            b.store_bit(0)
        else:
            b.store_bit(1)
            init = self.init.to_builder()
            can_inline_init = len(init.data) <= b.remaining_bits - 2 and len(init.refs) <= b.remaining_refs - 1
            if can_inline_init:
                b.store_bit(0)
                b.store_builder(init)
            else:
                b.store_bit(1)
                b.store_ref(init.end_cell())

        body = self.body.to_builder()
        can_inline_body = len(body.data) <= b.remaining_bits - 1 and len(body.refs) <= b.remaining_refs
        if can_inline_body:
            b.store_bit(0)
            b.store_builder(body)
        else:
            b.store_bit(1)
            b.store_ref(body.to_cell())
        return b
    
    @classmethod
    def internal(cls,
                 src: MsgAddressInt,
                 dest: MsgAddressInt,
                 value: int | CurrencyCollection,
                 body: Cell = Cell.empty(),
                 init: StateInit | None = None,
                 ihr_disabled: bool = True,
                 bounce: bool = True,
                 bounced: bool = False,
                 ihr_fee: int = 0,
                 fwd_fee: int = 0,
                 created_lt: int = 0,
                 created_at: int = 0,
                 ) -> Message:
        if isinstance(value, int):
            value = CurrencyCollection(value, ExtraCurrencyCollection())
    
        info = IntMsgInfo(
            ihr_disabled=ihr_disabled,
            bounce=bounce,
            bounced=bounced,
            src=src,
            dest=dest,
            value=value,
            ihr_fee=ihr_fee,
            fwd_fee=fwd_fee,
            created_lt=created_lt,
            created_at=created_at
        )
        return cls(info, init, body)

    @classmethod
    def external_out(cls,
                     src: MsgAddressInt,
                     body: Cell = Cell.empty(),
                     dest: MsgAddressExt = AddrNone(),
                     init: StateInit | None = None,
                     created_lt: int = 0,
                     created_at: int = 0) -> Message:
        info = ExtOutInfo(
            src=src,
            dest=dest,
            created_lt=created_lt,
            created_at=created_at
        )
        return cls(info, init, body)
    
    @classmethod
    def external_in(cls, *,
                     src: MsgAddressExt,
                     dest: MsgAddressInt,
                     body: Cell = Cell.empty(),
                     init: StateInit | None = None,
                     import_fee: int = 0) -> Message:
        info = ExtInInfo(
            src=src,
            dest=dest,
            import_fee=import_fee
        )
        return cls(info, init, body)

    def get_normalized_hash(self) -> bytes:
        if not isinstance(self.info, ExtInInfo):
            raise ValueError('Can calculate normalized hash only from External-in message')
        normalized_ext_in_info = ExtInInfo(
            src=AddrNone(),
            dest=self.info.dest,
            import_fee=0
        )
        c = (
            begin_cell()
            .store_tlb(normalized_ext_in_info)
            .store_maybe_ref(None)
            .store_bit(1) # Either bit
            .store_ref(self.body)
            .end_cell()
        )
        return c.hash()



from abc import ABC, abstractmethod
from typing import final, Self

from ..cell.cell import Cell
from ..cell.slice import Slice
from ..cell.builder import Builder
from ..common import BitString
from ..common.bitstring import int2bs

class TlbDeserializationError(Exception):
    pass

class TlbConstructor(ABC):
    @classmethod
    @abstractmethod
    def tag(cls) -> tuple[int, int] | None: ...

    @classmethod
    @abstractmethod
    def deserialize_fields(cls, s: Slice) -> Self: ...

    @abstractmethod
    def serialize_fields(self, b: Builder) -> Builder: ...

    @staticmethod
    def normalize_tag(tag: tuple[int, int] | None) -> BitString:
        if tag is None:
            return BitString()
        return int2bs(tag[0], tag[1])

    @final
    @classmethod
    def deserialize(cls, s: Slice) -> Self:
        try:
            s.skip_prefix(TlbConstructor.normalize_tag(cls.tag()))
        except Exception:
            raise TlbDeserializationError('Tag mismatch')
        return cls.deserialize_fields(s)

    @final
    def serialize(self, b: Builder | None = None) -> Builder:
        if b is None:
            b = Builder()
        b.store_bits(TlbConstructor.normalize_tag(self.tag()))
        self.serialize_fields(b)
        return b

    def to_builder(self) -> Builder:
        return self.serialize()

    def to_cell(self) -> Cell:
        return self.serialize().end_cell()

    def to_slice(self) -> Slice:
        return self.to_cell().begin_parse()

    @classmethod
    def from_slice(cls, s: Slice) -> Self:
        c = cls.deserialize(s)
        s.end_parse()
        return c

    @classmethod
    def from_cell(cls, c: Cell) -> Self:
        return cls.from_slice(c.begin_parse())



import base64

from typing import Literal

from .base_api_client import BaseApiClient
from .provider import Provider

from agton.ton import Cell, Slice, MsgAddressInt, Network
from agton.ton.types.tvm_value import TvmValue

def encode_tvm_value(v: TvmValue) -> str:
    if isinstance(v, int):
        return str(v)
        # return {
        #     'type': 'int257',
        #     'value': hex(v) 
        # }
    if isinstance(v, Cell):
        return v.to_boc().hex()
        # return {
        #     'type': 'cell',
        #     'value': v.to_boc().hex()
        # }
    if isinstance(v, Slice):
        return v.to_cell().to_boc().hex()
        # return {
        #     'type': 'slice',
        #     'value': v.to_cell().to_boc().hex()
        # }
    raise ValueError('Stack supports only int, Cell and Slice types')

def decode_tvm_value(d: dict) -> TvmValue:
    print(d)
    type_ = d.get('type')
    value = d.get('value')
    if type_ is None:
        raise ValueError('No type for tvm value')
    if value is None:
        raise ValueError('No value for tvm value')

    if type_ == 'num':
        return int(value, base=16)
    if type_ == 'cell':
        return Cell.from_boc(base64.b64decode(value))
    if type_ == 'slice':
        return Cell.from_boc(base64.b64decode(value)).to_slice()

    raise ValueError(f'Unexpected type for tvm value: {type_}')

class TonApiClient(Provider, BaseApiClient):
    def __init__(self, *,
                 net: Literal['testnet', 'mainnet'] = 'testnet',
                 api_key: str | None = None):
        host: str
        if net == 'mainnet':
            host = 'https://tonapi.io/'
        elif net == 'testnet':
            host = 'https://testnet.tonapi.io/'
        else:
            raise ValueError(f"Network should be 'mainnet' or 'testnet', but got {net}")

        BaseApiClient.__init__(self, host, api_key=('X-Api-Key', api_key) if api_key else None)


    def raw_run_get_method(self,
                        a: MsgAddressInt,
                        method_id: int,
                        stack: tuple[TvmValue, ...],
                        method: str | None = None) -> tuple[TvmValue, ...]:
        if method is None:
            raise ValueError('method_id as integer is not supported by TonApi')
        # data = {
        #     'args': [
        #         encode_tvm_value(v) for v in stack
        #     ]
        # }
        args = ''
        if stack:
            args = '?args=' + '&args='.join(encode_tvm_value(v) for v in stack)
            args += '&fix_order=true'
        url = f'/v2/blockchain/accounts/{a}/methods/{method}'
        if args:
            url += args
        print(self.base_url + url)
        r = self.post(url)
        print(r)
        s = r['stack']
        c = r['exit_code']
        if c != 0:
            raise ValueError(f'Non zero exit code during get method: {c}')
        return tuple(decode_tvm_value(v) for v in s)

    def raw_send_external_message(self, message: bytes):
        data = {
            'boc': message.hex()
        }
        return self.post('/v2/blockchain/message', json=data)



import nacl.signing
import nacl.exceptions

def private_key_to_public_key(private_key: bytes) -> bytes:
    return nacl.signing.SigningKey(private_key).verify_key.encode()

def sign(data: bytes, private_key: bytes) -> bytes:
    signer = nacl.signing.SigningKey(private_key)
    signed = signer.sign(data)
    return signed.signature

def verify(data: bytes, signature: bytes, public_key: bytes) -> bool:
    verifier = nacl.signing.VerifyKey(public_key)
    try:
        verifier.verify(data, signature)
        return True
    except nacl.exceptions.BadSignatureError:
        return False



from typing import Iterable

class BytesParser:
    def __init__(self, b: bytes):
        self.b = b
        self.i = 0
    
    def preload_bytes(self, n: int) -> bytes:
        return self.b[self.i:self.i + n]
    
    def load_bytes(self, n: int) -> bytes:
        self.i += n
        return self.b[self.i - n:self.i]
    
    def skip_bytes(self, n: int):
        self.i += n

    def expect(self, v: int | Iterable[int] | bytes):
        if isinstance(v, int):
            v = [v]
        if not isinstance(v, bytes):
            v = bytes(v)
        n = len(v)
        b = self.preload_bytes(n)
        if b != v:
            raise ValueError('Expected {v}, but {b} found')
        self.i += n
    
    def load_uint(self, byte_size: int) -> int:
        b = self.load_bytes(byte_size)
        return int.from_bytes(b, 'big')
    
    def end_parse(self):
        if self.i != len(self.b):
            raise ValueError('Expected end of bytes')



from __future__ import annotations

import base64

from typing import Self
from bitarray import bitarray
from dataclasses import dataclass

from agton.ton.common.bitstring import BitString

from ..cell.slice import Slice
from ..cell.builder import Builder
from .tlb import TlbConstructor, TlbDeserializationError

from ..crypto import crc16


@dataclass(frozen=True, slots=True)
class AddrNone(TlbConstructor):
    '''addr_none$00 = MsgAddressExt'''

    @classmethod
    def tag(cls):
        return 0b00, 2

    def serialize_fields(self, b: Builder) -> Builder:
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        return cls()


@dataclass(frozen=True, slots=True)
class AddrExtern(TlbConstructor):
    '''
    addr_extern$01 len:(## 9) external_address:(bits len) 
    = MsgAddressExt;
    '''
    len: int
    external_address: BitString

    @classmethod
    def tag(cls):
        return 0b01, 2
    
    def serialize_fields(self, b: Builder) -> Builder:
        b.store_uint(self.len, 9)
        b.store_bits(self.external_address)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        len = s.load_uint(9)
        external_address = s.load_bits(len)
        return cls(len, external_address)

@dataclass(frozen=True, slots=True)
class Anycast(TlbConstructor):
    '''
    anycast_info$_ depth:(#<= 30) { depth >= 1 }
    rewrite_pfx:(bits depth) = Anycast;
    '''
    depth: int
    rewrite_pfx: bitarray

    @classmethod
    def tag(cls):
        return None

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_uint(self.depth, 5)
        b.store_bits(self.rewrite_pfx)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        depth = s.load_uint(5)
        if not (1 <= depth <= 30):
            raise TlbDeserializationError()
        rewrite_pfx = s.load_bits(depth)
        return cls(depth, rewrite_pfx)

PAYLOAD_LEN = 36
FLAG_TESTNET = 0x80
FLAG_BOUNCEABLE = 0x11
FLAG_NON_BOUNCEABLE = 0x51

@dataclass(frozen=True, slots=True)
class AddressFlags:
    bounceable: bool
    testnet_only: bool

@dataclass(frozen=True, slots=True)
class AddrStd(TlbConstructor):
    '''
    addr_std$10 anycast:(Maybe Anycast) 
    workchain_id:int8 address:bits256 = MsgAddressInt;
    '''
    workchain: int
    address: bytes
    anycast: Anycast | None = None

    @classmethod
    def tag(cls):
        return 0b10, 2

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_maybe_tlb(self.anycast)
        b.store_int(self.workchain, 8)
        b.store_bytes(self.address)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        anycast = s.load_maybe_tlb(Anycast)
        workchain = s.load_int(8)
        address = s.load_bytes(32)
        return cls(workchain, address, anycast)
    
    class ParseError(ValueError):
        pass

    @classmethod
    def parse_with_flags(cls, s: str) -> tuple[Self, AddressFlags]:
        if not isinstance(s, str) or not s:
            raise cls.ParseError("Address string is empty or not a string.")

        s = s.replace("-", "+").replace("_", "/")
        if len(s) % 4:
            s += "=" * (4 - (len(s) % 4))

        try:
            b = base64.b64decode(s, validate=True)
        except Exception as e:
            raise cls.ParseError(f"Invalid base64 string: {e}") from e

        if len(b) != PAYLOAD_LEN:
            raise cls.ParseError(
                f"Human readable form must be {PAYLOAD_LEN} bytes, got {len(b)}."
            )

        flags = b[0]
        wc = int.from_bytes(b[1:2], signed=True)
        addr = b[2:34]
        checksum = b[34:36]

        testnet_only = bool(flags & FLAG_TESTNET)
        flags &= ~FLAG_TESTNET

        if flags == FLAG_BOUNCEABLE:
            bounceable = True
        elif flags == FLAG_NON_BOUNCEABLE:
            bounceable = False
        else:
            raise cls.ParseError("Invalid address flags.")

        expected = crc16(b[:34])
        if expected != checksum:
            raise cls.ParseError("Invalid checksum.")

        return cls(wc, addr), AddressFlags(bounceable, testnet_only)

    @classmethod
    def parse(cls, s: str) -> Self:
        a, _ = cls.parse_with_flags(s)
        return a
    
    def raw(self) -> str:
        return f'{self.workchain}:{self.address.hex().zfill(32)}'
    
    def format(self, *,
               raw: bool = False,
               urlsafe: bool = True,
               bounceable: bool = True,
               testnet_only: bool = False) -> str:
        if raw:
            return self.raw()
        f = FLAG_BOUNCEABLE if bounceable else FLAG_NON_BOUNCEABLE
        if testnet_only:
            f |= FLAG_TESTNET

        buf = bytearray(PAYLOAD_LEN)
        buf[0] = f
        buf[1:2] = self.workchain.to_bytes(1, "big", signed=True)
        buf[2:34] = self.address
        buf[34:36] = crc16(buf[:34])

        b64_func = base64.urlsafe_b64encode if urlsafe else base64.b64encode
        encoded = b64_func(bytes(buf)).decode()

        return encoded
    
    def __format__(self, spec: str) -> str:
        raw = "r" in spec
        bounceable = "n" not in spec
        testnet_only = "t" in spec
        return self.format(
            raw=raw, urlsafe=True, bounceable=bounceable, testnet_only=testnet_only
        )

    def __repr__(self) -> str:
        return f'Address({self.raw()})'
    
    def __str__(self) -> str:
        return self.format()


@dataclass(frozen=True, slots=True)
class AddrVar(TlbConstructor):
    '''
    addr_var$11 anycast:(Maybe Anycast) addr_len:(## 9) 
    workchain_id:int32 address:(bits addr_len) = MsgAddressInt;
    '''
    anycast: Anycast | None
    len: int
    workchain: int
    address: bitarray

    @classmethod
    def tag(cls):
        return 0b11, 2

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_maybe_tlb(self.anycast)
        b.store_int(self.len, 9)
        b.store_int(self.workchain, 32)
        b.store_bits(self.address)
        return b

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        anycast = s.load_maybe_tlb(Anycast)
        len = s.load_uint(9)
        workchain = s.load_int(8)
        address = s.load_bits(len)
        return cls(anycast, len, workchain, address)

MsgAddressExt = AddrNone | AddrExtern
MsgAddressInt = AddrStd | AddrVar
MsgAddress = MsgAddressExt | MsgAddressInt

Address = AddrStd

def msg_address(s: Slice) -> MsgAddress:
    tag = s.preload_uint(2) 
    match tag:
        case 0b00: return AddrNone.deserialize(s)
        case 0b01: return AddrExtern.deserialize(s)
        case 0b10: return AddrStd.deserialize(s)
        case 0b11: return AddrVar.deserialize(s)
    assert False


def msg_address_ext(s: Slice) -> MsgAddressExt:
    tag = s.preload_uint(2) 
    match tag:
        case 0b00: return AddrNone.deserialize(s)
        case 0b01: return AddrExtern.deserialize(s)
    raise TlbDeserializationError()


def msg_address_int(s: Slice) -> MsgAddressInt:
    tag = s.preload_uint(2) 
    match tag:
        case 0b10: return AddrStd.deserialize(s)
        case 0b11: return AddrVar.deserialize(s)
    raise TlbDeserializationError()



from __future__ import annotations

from typing import Callable, Self
from dataclasses import dataclass, field
from ..common import BitString, int2bs, bs2int

from ..cell import Slice
from ..cell import Builder, begin_cell
from ..cell import Cell

from .msg_address import MsgAddress, Address, msg_address

from .tlb import TlbDeserializationError

'''
hm_edge#_ {n:#} {X:Type} {l:#} {m:#} label:(HmLabel ~l n) 
          {n = (~m) + l} node:(HashmapNode m X) = Hashmap n X;

hmn_leaf#_ {X:Type} value:X = HashmapNode 0 X;
hmn_fork#_ {n:#} {X:Type} left:^(Hashmap n X) 
           right:^(Hashmap n X) = HashmapNode (n + 1) X;

hml_short$0 {m:#} {n:#} len:(Unary ~n) {n <= m} s:(n * Bit) = HmLabel ~n m;
hml_long$10 {m:#} n:(#<= m) s:(n * Bit) = HmLabel ~n m;
hml_same$11 {m:#} v:Bit n:(#<= m) = HmLabel ~n m;

unary_zero$0 = Unary ~0;
unary_succ$1 {n:#} x:(Unary ~n) = Unary ~(n + 1);

hme_empty$0 {n:#} {X:Type} = HashmapE n X;
hme_root$1 {n:#} {X:Type} root:^(Hashmap n X) = HashmapE n X;
'''

@dataclass
class Hashmap:
    label: BitString
    node: Leaf | Fork

    def to_dict(self) -> dict[BitString, Slice]:
        d = dict()
        match self.node:
            case Leaf(s):
                d[self.label] = s
                return d
            case Fork(l, r):
                left = l.to_dict()
                right = r.to_dict()
                for k, v in left.items():
                    d[self.label + BitString([0]) + k] = v
                for k, v in right.items():
                    d[self.label + BitString([1]) + k] = v
                return d

    @classmethod
    def from_dict(cls, d: dict[BitString, Slice]) -> Hashmap:
        def lcp(a: list[BitString]) -> int:
            # todo: Benchmark and maybe optimize, currently O(n * m)
            n = len(a)
            m = len(a[0])
            for i in range(m):
                b = a[0][i]
                for j in range(n):
                    if b != a[j][i]:
                        return i
            return n

        keys = list(d.keys())
        if len(keys) == 1:
            k = keys[0]
            return Hashmap(k, Leaf(d[k]))
        k = lcp(keys)
        label = keys[0][:k]

        ld: dict[BitString, Slice] = dict()
        rd: dict[BitString, Slice] = dict()
        for key in keys:
            if key[k] == 0:
                ld[BitString(key[k + 1:])] = d[key]
            elif key[k] == 1:
                rd[BitString(key[k + 1:])] = d[key]
        l = cls.from_dict(ld)
        r = cls.from_dict(rd)

        return cls(BitString(label), Fork(l, r))

HashmapE = Hashmap | None

@dataclass
class Leaf:
    value: Slice

@dataclass
class Fork:
    left: Hashmap
    right: Hashmap

def load_unary(s: Slice) -> int:
    ans = 0
    while s.load_bit() == 1:
        ans += 1
    return ans

def store_unary(b: Builder, n: int):
    for _ in range(n):
        b.store_bit(1)
    b.store_bit(0)

def load_hml_short(s: Slice, m: int) -> tuple[BitString, int]:
    s.skip_prefix(int2bs(0b0, 1))
    n = load_unary(s)
    if not (n <= m):
        raise TlbDeserializationError()
    return s.load_bits(n), n

def load_hml_long(s: Slice, m: int) -> tuple[BitString, int]:
    s.skip_prefix(int2bs(0b10, 2))
    n = s.load_uint(m.bit_length())
    return s.load_bits(n), n

def load_hml_same(s: Slice, m: int) -> tuple[BitString, int]:
    s.skip_prefix(int2bs(0b11, 2))
    v = s.load_bit()
    n = s.load_uint(m.bit_length())
    return BitString([v] * n), n

def load_label(s: Slice, m: int) -> tuple[BitString, int]:
    tag = s.preload_uint(2)
    match tag:
        case 0b00 | 0b01: return load_hml_short(s, m)
        case 0b10: return load_hml_long(s, m)
        case 0b11: return load_hml_same(s, m)
    assert False

def store_label(b: Builder, l: BitString, m: int):
    raise NotImplementedError

def load_node(s: Slice, n: int) -> Leaf | Fork:
    if n == 0:
        return Leaf(s)
    left = load_hashmap(s.load_ref().begin_parse(), n - 1)
    right = load_hashmap(s.load_ref().begin_parse(), n - 1)
    return Fork(left, right)

def store_node(b: Builder, n: int, node: Leaf | Fork):
    raise NotImplementedError

def load_hashmap(s: Slice, n: int) -> Hashmap:
    label, l = load_label(s, n)
    m = n - l
    node = load_node(s, m)
    return Hashmap(label, node)

def store_hashmap(b: Builder, h: Hashmap) -> Builder:
    raise NotImplementedError

@dataclass(frozen=True, slots=True)
class HashmapCodec[K, V]:
    k_de: Callable[[BitString], K] | None = None
    k_se: Callable[[K], BitString] | None = None
    v_de: Callable[[Slice], V] | None = None
    v_se: Callable[[V], Slice] | None = None

    def decode(self, hashmap: HashmapE) -> dict[K, V]:
        if self.k_de is None or self.v_de is None:
            raise ValueError('Deserializators are not set')
        d = {} if hashmap is None else hashmap.to_dict()
        return {self.k_de(k): self.v_de(v) for k, v in d.items()}
    
    def encode(self, d: dict[K, V]) -> HashmapE:
        if not d:
            return None
        if self.k_se is None or self.v_se is None:
            raise ValueError('Serializators are not set')
        cd = {self.k_se(k): self.v_se(v) for k, v in d.items()}
        hashmap = Hashmap.from_dict(cd)
        return hashmap
    
    def with_int_values(self, n: int) -> HashmapCodec[K, int]:
        def v_se(v: int) -> Slice:
            return begin_cell().store_int(v, n).to_slice()
        def v_de(s: Slice) -> int:
            return s.load_int(n)
        return HashmapCodec(self.k_de, self.k_se, v_de, v_se)
    
    def with_uint_values(self, n: int) -> HashmapCodec[K, int]:
        def v_se(v: int) -> Slice:
            return begin_cell().store_uint(v, n).to_slice()
        def v_de(s: Slice) -> int:
            return s.load_uint(n)
        return HashmapCodec(self.k_de, self.k_se, v_de, v_se)
    
    def with_var_uint_values(self, n: int) -> HashmapCodec[K, int]:
        def v_se(v: int) -> Slice:
            return begin_cell().store_var_uint(v, n).to_slice()
        def v_de(s: Slice) -> int:
            return s.load_var_uint(n)
        return HashmapCodec(self.k_de, self.k_se, v_de, v_se)
    
    def with_coins_values(self) -> HashmapCodec[K, int]:
        def v_se(v: int) -> Slice:
            return begin_cell().store_coins(v).to_slice()
        def v_de(s: Slice) -> int:
            return s.load_coins()
        return HashmapCodec(self.k_de, self.k_se, v_de, v_se)
    
    def with_snake_data_values(self) -> HashmapCodec[K, str]:
        raise NotImplementedError
    
    def with_msg_address_values(self) -> HashmapCodec[K, MsgAddress]:
        def v_se(v: MsgAddress) -> Slice:
            return begin_cell().store_tlb(v).to_slice()
        def v_de(s: Slice) -> MsgAddress:
            return s.load_msg_address()
        return HashmapCodec(self.k_de, self.k_se, v_de, v_se)
    
    def with_address_values(self) -> HashmapCodec[K, Address]:
        def v_se(v: Address) -> Slice:
            return begin_cell().store_tlb(v).to_slice()
        def v_de(s: Slice) -> Address:
            return s.load_tlb(Address)
        return HashmapCodec(self.k_de, self.k_se, v_de, v_se)
    
    def with_int_keys(self, n: int) -> HashmapCodec[int, V]:
        def k_se(v: int) -> BitString:
            return begin_cell().store_int(v, n).to_cell().data
        def k_de(b: BitString) -> int:
            return bs2int(b, signed=True)
        return HashmapCodec(k_de, k_se, self.v_de, self.v_se)
    
    def with_uint_keys(self, n: int) -> HashmapCodec[int, V]:
        def k_se(v: int) -> BitString:
            return begin_cell().store_uint(v, n).to_cell().data
        def k_de(b: BitString) -> int:
            return bs2int(b, signed=False)
        return HashmapCodec(k_de, k_se, self.v_de, self.v_se)
    
    def with_address_keys(self) -> HashmapCodec[Address, V]:
        def k_se(a: Address) -> BitString:
            return begin_cell().store_address(a).to_cell().data
        def k_de(b: BitString) -> Address:
            return Cell(b).begin_parse().load_address()
        return HashmapCodec(k_de, k_se, self.v_de, self.v_se)



from __future__ import annotations

from dataclasses import dataclass
from typing import Self, Type

from ..cell.slice import Slice
from ..cell.builder import Builder

from .tlb import TlbConstructor

from .msg_address import MsgAddressInt, MsgAddressExt
from .currency_collection import CurrencyCollection


@dataclass(frozen=True, slots=True)
class IntMsgInfo(TlbConstructor):
    '''
    int_msg_info$0 ihr_disabled:Bool bounce:Bool bounced:Bool
    src:MsgAddressInt dest:MsgAddressInt 
    value:CurrencyCollection ihr_fee:Grams fwd_fee:Grams
    created_lt:uint64 created_at:uint32 = CommonMsgInfo;
    '''
    ihr_disabled: bool
    bounce: bool
    bounced: bool
    src: MsgAddressInt
    dest: MsgAddressInt
    value: CurrencyCollection
    ihr_fee: int
    fwd_fee: int
    created_lt: int
    created_at: int

    @classmethod
    def tag(cls):
        return 0b0, 1
    
    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_bool(self.ihr_disabled)
            .store_bool(self.bounce)
            .store_bool(self.bounced)
            .store_tlb(self.src)
            .store_tlb(self.dest)
            .store_tlb(self.value)
            .store_coins(self.ihr_fee)
            .store_coins(self.fwd_fee)
            .store_uint(self.created_lt, 64)
            .store_uint(self.created_at, 32)
        )

    @classmethod
    def deserialize_fields(cls, s: Slice) -> IntMsgInfo:
        ihr_disabled = s.load_bool()
        bounce = s.load_bool()
        bounced = s.load_bool()
        src = s.load_msg_address_int()
        dest = s.load_msg_address_int()
        value = s.load_tlb(CurrencyCollection)
        ihr_fee = s.load_coins()
        fwd_fee = s.load_coins()
        created_lt = s.load_uint(64)
        created_at = s.load_uint(32)

        return cls(ihr_disabled, bounce, bounced, src, dest, value, ihr_fee, fwd_fee, created_lt, created_at)


@dataclass(frozen=True, slots=True)
class ExtInInfo(TlbConstructor):
    '''
    ext_in_msg_info$10 src:MsgAddressExt dest:MsgAddressInt 
    import_fee:Grams = CommonMsgInfo;
    '''

    src: MsgAddressExt
    dest: MsgAddressInt
    import_fee: int

    @classmethod
    def tag(cls):
        return 0b10, 2

    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_tlb(self.src)
            .store_tlb(self.dest)
            .store_coins(self.import_fee)
        )

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        src = s.load_msg_address_ext()
        dest = s.load_msg_address_int()
        import_fee = s.load_coins()
        return cls(src, dest, import_fee)


@dataclass(frozen=True, slots=True)
class ExtOutInfo(TlbConstructor):
    '''
    ext_out_msg_info$11 src:MsgAddressInt dest:MsgAddressExt
        created_lt:uint64 created_at:uint32 = CommonMsgInfo;
    '''

    src: MsgAddressInt
    dest: MsgAddressExt
    created_lt: int
    created_at: int

    @classmethod
    def tag(cls):
        return 0b11, 2

    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_tlb(self.src)
            .store_tlb(self.dest)
            .store_uint(self.created_lt, 64)
            .store_uint(self.created_at, 32)
        )

    @classmethod
    def deserialize_fields(cls, s: Slice) -> ExtOutInfo:
        src = s.load_msg_address_int()
        dest = s.load_msg_address_ext()
        created_lt = s.load_uint(64)
        create_at = s.load_uint(32)
        return cls(src, dest, created_lt, create_at)

CommonMsgInfo = IntMsgInfo | ExtInInfo | ExtOutInfo

def common_msg_info(s: Slice) -> CommonMsgInfo:
    tag = s.preload_uint(2)
    match tag:
        case 0b00 | 0b01: return IntMsgInfo.deserialize(s)
        case 0b10: return ExtInInfo.deserialize(s)
        case 0b11: return ExtOutInfo.deserialize(s)
    assert False



from .cell.cell import Cell
from .cell.builder import Builder, begin_cell
from .cell.slice import Slice

from .types.network import Network

from .types.tlb import TlbConstructor
from .types import Address, MsgAddress, MsgAddressInt, MsgAddressExt
from .types import Message, MessageRelaxed
from .types import Hashmap, HashmapE, HashmapCodec
from .types import CurrencyCollection
from .types import StateInit

from .provider import ToncenterClient, TonApiClient, Provider
from .contracts.contract import Contract

from .utils import to_nano, from_nano, to_units, from_units, comment



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable

from agton.ton import Cell
from agton.ton.types.tvm_value import TvmValue
from agton.ton.crypto import crc16
from agton.ton import MsgAddressInt, Message


class ProviderError(Exception):
    pass


class Provider(ABC):
    @abstractmethod
    def raw_run_get_method(
        self, 
        a: MsgAddressInt,
        method_id: int,
        stack: tuple[TvmValue, ...],
        method: str | None = None
    ) -> tuple[TvmValue, ...]: ...

    @abstractmethod
    def raw_send_external_message(self, message: bytes): ...

    def _normalize_message(self, message: Message | Cell | bytes) -> Message:
        if isinstance(message, (bytes, bytearray, memoryview)):
            try:
                message = Cell.from_boc(message)
            except Exception as e:
                raise ValueError("Message is incorrectly serialized") from e
        if isinstance(message, Cell):
            try:
                message = Message.from_cell(message)
            except Exception as e:
                raise ValueError("Message is incorrectly serialized") from e
        return message
    
    def _normalize_stack(self, stack: Iterable[TvmValue] | TvmValue | None = None) -> tuple[TvmValue, ...]:
        if stack is None:
            return ()
        if isinstance(stack, Iterable):
            return tuple(stack)
        return (stack,)
    
    def send_external_message(self, message: Message | Cell | bytes) -> bytes:
        message = self._normalize_message(message)
        h = message.get_normalized_hash()
        self.raw_send_external_message(message.to_cell().to_boc())
        return h

    def run_get_method(self,
                       a: MsgAddressInt,
                       method_id: int | str,
                       stack: Iterable[TvmValue] | TvmValue | None = None) -> tuple[TvmValue, ...] | TvmValue:
        method_name_fallback = None if isinstance(method_id, int) else method_id
        if isinstance(method_id, str):
            t = int.from_bytes(crc16(method_id.encode()), byteorder='big')
            method_id = (t & 0xffff) | 0x10000
        stack = self._normalize_stack(stack)
        s = self.raw_run_get_method(a, method_id, stack, method_name_fallback)

        if len(s) == 1:
            return s[0]
        return s



from bitarray import frozenbitarray
from bitarray.util import int2ba, ba2int

BitString = frozenbitarray

def int2bs(v: int, len: int) -> BitString:
    return BitString(int2ba(v, len))

def bs2int(bs: BitString, signed: bool = True) -> int:
    return ba2int(bs, signed=signed)


from .msg_address import Address
from .msg_address import AddrExtern, AddrNone, AddrStd, AddrVar
from .msg_address import MsgAddress, MsgAddressInt, MsgAddressExt

from .common_msg_info import CommonMsgInfo, IntMsgInfo, ExtInInfo, ExtOutInfo
from .common_msg_info_relaxed import CommonMsgInfoRelaxed, IntMsgInfoRelaxed, ExtOutInfoRelaxed

from .currency_collection import CurrencyCollection, ExtraCurrencyCollection

from .state_init import StateInit

from .message import Message
from .message_relaxed import MessageRelaxed

from .hashmap import Hashmap, HashmapE, HashmapCodec

__all__ = [
    'Address',
    'AddrExtern',
    'AddrNone',
    'AddrStd',
    'AddrVar',
    'MsgAddress',
    'MsgAddressInt',
    'MsgAddressExt',
    'IntMsgInfo',
    'ExtInInfo',
    'ExtOutInfo',
    'CommonMsgInfo',
    'IntMsgInfoRelaxed',
    'ExtOutInfoRelaxed',
    'CommonMsgInfoRelaxed',
    'CurrencyCollection',
    'ExtraCurrencyCollection',
    'StateInit',
    'Message',
    'MessageRelaxed',
    'Hashmap', 
    'HashmapE', 
    'HashmapCodec'
]



import base64

from typing import Literal

from .base_api_client import BaseApiClient
from .provider import Provider

from agton.ton import Cell, Slice, MsgAddressInt, Network
from agton.ton.types.tvm_value import TvmValue

class ToncenterError(Exception):
    pass

def encode_tvm_value(v: TvmValue) -> dict:
    if isinstance(v, int):
        return {
            'type': 'num',
            'value': hex(v) 
        }
    if isinstance(v, Cell):
        return {
            'type': 'cell',
            'value': base64.b64encode(v.to_boc()).decode()
        }
    if isinstance(v, Slice):
        return {
            'type': 'slice',
            'value': base64.b64encode(v.to_cell().to_boc()).decode()
        }
    raise ToncenterError('Stack supports only int, Cell and Slice types')

def decode_tvm_value(d: dict) -> TvmValue:
    type_ = d.get('type')
    value = d.get('value')
    if type_ is None:
        raise ToncenterError('No type for tvm value')
    if value is None:
        raise ToncenterError('No value for tvm value')

    if type_ == 'num':
        return int(value, base=16)
    if type_ == 'cell':
        return Cell.from_boc(base64.b64decode(value))
    if type_ == 'slice':
        return Cell.from_boc(base64.b64decode(value)).to_slice()

    raise ToncenterError(f'Unexpected type for tvm value: {type_}')
    

class ToncenterClient(Provider, BaseApiClient):
    def __init__(self, *,
                 net: Literal['testnet', 'mainnet'] = 'testnet',
                 api_key: str | None = None):
        host: str
        if net == 'mainnet':
            host = 'https://toncenter.com/api/v3/'
        elif net == 'testnet':
            host = 'https://testnet.toncenter.com/api/v3/'
        else:
            raise ToncenterError(f"Network should be 'mainnet' or 'testnet', but got {net}")
        
        BaseApiClient.__init__(self, host, api_key=('X-Api-Key', api_key) if api_key else None)


    def raw_run_get_method(self,
                        a: MsgAddressInt,
                        method_id: int,
                        stack: tuple[TvmValue, ...],
                        method: str | None = None) -> tuple[TvmValue, ...]:
        if method is None:
            raise ToncenterError('method_id as integer is not supported by Toncenter')
        data = {
            'address': str(a),
            'method': method,
            'stack': [
                encode_tvm_value(v) for v in stack
            ]
        }
        r = self.post('/runGetMethod', json=data)
        s = r['stack']
        c = r['exit_code']
        if c != 0:
            raise ToncenterError(f'Non zero exit code during get method: {c}')
        return tuple(decode_tvm_value(v) for v in s)

    def raw_send_external_message(self, message: bytes):
        data = {
            'boc': base64.b64encode(message).decode()
        }
        return self.post('/message', json=data)



from decimal import Decimal, ROUND_DOWN

from .cell.cell import Cell
from .cell.builder import begin_cell

def to_units(value: str | float | int, decimals: int) -> int:
    try:
        tons = Decimal(str(value))
        nanounits = (tons * 10**decimals).to_integral_value(rounding=ROUND_DOWN)
        return int(nanounits)
    except Exception as e:
        raise ValueError(f"Invalid input for conversion: {value}") from e

def from_units(units: int, decimals: int) -> str:
    try:
        tons = Decimal(units) / 10**decimals
        return str(tons.normalize())
    except Exception as e:
        raise ValueError(f"Invalid input for conversion: {units}") from e

def to_nano(value: str | float | int) -> int:
    return to_units(value, 9)

def from_nano(units: int) -> str:
    return from_units(units, 9)

def comment(s: str) -> Cell:
    return (
        begin_cell()
        .store_uint(0, 32)
        .store_snake_string(s)
        .end_cell()
    )
